/***************************************************************************
   NS8390 Device description header

   Header file for the National Semiconductor NS8390 Ethernet chipset, 
   commonly used in PC network cards like NE2000, WD80x3 and 3c503.

   This file was compiled from the National Semicondutor datasheet for the
   NS83907, Motorola SBC5206 dBUG monitor source and the header for XXXs
   WD80x3 pc386 driver.

   Author:
     David Fiddes, D.J.Fiddes@hw.ac.uk
     http://www.calm.hw.ac.uk/davidf/coldfire/
 
   COPYRIGHT (c) 1989-1998.
   On-Line Applications Research Corporation (OAR).
   Copyright assigned to U.S. Government, 1994.
 
   The license and distribution terms for this file may be
   found in the file LICENSE in this distribution or at
 
   http://www.OARcorp.com/rtems/license.html.
  
   $Id: ns8390.h,v 1.11 1998/02/19 23:02:02 joel Exp $
 ***************************************************************************/

#ifndef _NS8390_H
#define _NS8390_H

#include <ka9q/rtems_ka9q.h>
#include <ka9q/global.h>
#include <ka9q/enet.h>
#include <ka9q/iface.h>
#include <ka9q/netuser.h>
#include <ka9q/trace.h>
#include <ka9q/commands.h>
#include <ka9q/domain.h>

#include <stats.h>

/*
 * Memory usage defines
 */
#define NUM_TX_PAGES	6 /* a single packet for non pipelined transmits */


/*
 *  Types for get and set register routines
 */

typedef unsigned8 (*getRegister_f)(unsigned32 port, unsigned8 register);
typedef void      (*setRegister_f)(
                            unsigned32 port, unsigned8 reg, unsigned8 value);
typedef unsigned16	(*getData_f)(unsigned32 port);
typedef void		(*setData_f)(unsigned32 port, unsigned16 value);


/*
 *  Types for ns8390 user provided routines
 */

/* useful forward declarations */
struct ns8390_device;
struct ns8390_pkt_hdr;

typedef void (*reset_ns8390_f)		(struct ns8390_device *);
typedef	void (*get_ns8390_hdr_f)	(struct ns8390_device *, struct ns8390_pkt_hdr *, int);
typedef void (*block_output_f)		(struct ns8390_device *, int *, struct mbuf **, int);
typedef void (*block_input_f)		(struct ns8390_device *, int, struct mbuf *, int);



/* The 8390 specific per-packet-header format. */
struct ns8390_pkt_hdr {
  unsigned char status; /* status */
  unsigned char next;   /* pointer to next packet. */
  unsigned short count; /* header + packet length in bytes */
};

struct ns8390_device {
	struct iface		*iface;
	rtems_id			txWaitTid;

	/* NS8390 buffer management  - suspect */
	struct mbuf			**rxMbuf;
	struct mbuf			**txMbuf;
	unsigned int 		port;
	char 				*base;
	unsigned long		bpar;
	unsigned int		irno;
	int					tx_start_page;
	int					rx_start_page;
	int					stop_page;
	int					current_page;

	unsigned16			tx1;
	unsigned16			tx2;

	/* client mini-port driver functions */
	reset_ns8390_f		reset_ns8390;
	get_ns8390_hdr_f	get_ns8390_hdr;
	block_output_f		block_output;
	block_input_f		block_input;

	/* chip access functions */
	getRegister_f  getRegister;
	setRegister_f  setRegister;
	getData_f      getData;
	setData_f      setData;

	/* standard ethernet statistics */
	eth_statistics eth_stats;

	unsigned long	txRawWait;
};


/* Exported functions in ns8390.c */
int
ns8390_driver_attach ( struct ns8390_device **user_dp );

void
ns8390_init( struct ns8390_device *dp );

void
ns8390_start( struct ns8390_device *dp );

rtems_isr
ns8390_interrupt_handler (rtems_vector_number v);



/* Data port through which all "dma" access takes place */

#define NS8390_PAGE0_DATAPORT		(0x10)


/* PAGE0 Read registers */
#define NS8390_PAGE0_READ_CR		(0x00)
#define NS8390_PAGE0_READ_CLDA0		(0x01)
#define NS8390_PAGE0_READ_CLDA1		(0x02)
#define NS8390_PAGE0_READ_BNRY		(0x03)
#define NS8390_PAGE0_READ_TSR		(0x04)
#define NS8390_PAGE0_READ_NCR		(0x05)
#define NS8390_PAGE0_READ_FIFO		(0x06)
#define NS8390_PAGE0_READ_ISR		(0x07)
#define NS8390_PAGE0_READ_CRDA0		(0x08)
#define NS8390_PAGE0_READ_CRDA1		(0x09)
#define NS8390_PAGE0_READ_reserved0	(0x0A)
#define NS8390_PAGE0_READ_reserved1	(0x0B)
#define NS8390_PAGE0_READ_RSR		(0x0C)
#define NS8390_PAGE0_READ_CNTR0		(0x0D)
#define NS8390_PAGE0_READ_CNTR1		(0x0E)
#define NS8390_PAGE0_READ_CNTR2		(0x0F)

/* PAGE0 Write registers */
#define NS8390_PAGE0_WRITE_CR		(0x00)
#define NS8390_PAGE0_WRITE_PSTART	(0x01)
#define NS8390_PAGE0_WRITE_PSTOP	(0x02)
#define NS8390_PAGE0_WRITE_BNRY		(0x03)
#define NS8390_PAGE0_WRITE_TPSR		(0x04)
#define NS8390_PAGE0_WRITE_TBCR0	(0x05)
#define NS8390_PAGE0_WRITE_TBCR1	(0x06)
#define NS8390_PAGE0_WRITE_ISR		(0x07)
#define NS8390_PAGE0_WRITE_RSAR0	(0x08)
#define NS8390_PAGE0_WRITE_RSAR1	(0x09)
#define NS8390_PAGE0_WRITE_RBCR0	(0x0A)
#define NS8390_PAGE0_WRITE_RBCR1	(0x0B)
#define NS8390_PAGE0_WRITE_RCR		(0x0C)
#define NS8390_PAGE0_WRITE_TCR		(0x0D)
#define NS8390_PAGE0_WRITE_DCR		(0x0E)
#define NS8390_PAGE0_WRITE_IMR		(0x0F)


/* PAGE1 Read/Write registers */
#define NS8390_PAGE1_CR				(0x00)
#define NS8390_PAGE1_PAR0			(0x01)
#define NS8390_PAGE1_PAR1			(0x02)
#define NS8390_PAGE1_PAR2			(0x03)
#define NS8390_PAGE1_PAR3			(0x04)
#define NS8390_PAGE1_PAR4			(0x05)
#define NS8390_PAGE1_PAR5			(0x06)
#define NS8390_PAGE1_CURR			(0x07)
#define NS8390_PAGE1_MAR0			(0x08)
#define NS8390_PAGE1_MAR1			(0x09)
#define NS8390_PAGE1_MAR2			(0x0A)
#define NS8390_PAGE1_MAR3			(0x0B)
#define NS8390_PAGE1_MAR4			(0x0C)
#define NS8390_PAGE1_MAR5			(0x0D)
#define NS8390_PAGE1_MAR6			(0x0E)
#define NS8390_PAGE1_MAR7			(0x0F)


/* PAGE2 Read registers */
#define NS8390_PAGE2_READ_CR		(0x00)
#define NS8390_PAGE2_READ_PSTART	(0x01)
#define NS8390_PAGE2_READ_PSTOP		(0x02)
#define NS8390_PAGE2_READ_RNPP		(0x03)
#define NS8390_PAGE2_READ_TPSR		(0x04)
#define NS8390_PAGE2_READ_LNPP		(0x05)
#define NS8390_PAGE2_READ_ACU		(0x06)
#define NS8390_PAGE2_READ_ACL		(0x07)
#define NS8390_PAGE2_READ_reserved0	(0x08)
#define NS8390_PAGE2_READ_reserved1	(0x09)
#define NS8390_PAGE2_READ_reserved2 (0x0A)
#define NS8390_PAGE2_READ_reserved3 (0x0B)
#define NS8390_PAGE2_READ_RCR		(0x0C)
#define NS8390_PAGE2_READ_TCR		(0x0D)
#define NS8390_PAGE2_READ_DCR		(0x0E)
#define NS8390_PAGE2_READ_IMR		(0x0F)

/* PAGE2 Write registers */
#define NS8390_PAGE2_WRITE_CR		(0x00)
#define NS8390_PAGE2_WRITE_CLDA0	(0x01)
#define NS8390_PAGE2_WRITE_CLDA1	(0x02)
#define NS8390_PAGE2_WRITE_RNPP		(0x03)
#define NS8390_PAGE2_WRITE_reservd0	(0x04)
#define NS8390_PAGE2_WRITE_LNPP		(0x05)
#define NS8390_PAGE2_WRITE_ACU		(0x06)
#define NS8390_PAGE2_WRITE_ACL		(0x07)
#define NS8390_PAGE2_WRITE_reservd1	(0x08)
#define NS8390_PAGE2_WRITE_reservd2	(0x09)
#define NS8390_PAGE2_WRITE_reservd3	(0x0A)
#define NS8390_PAGE2_WRITE_reservd4	(0x0B)
#define NS8390_PAGE2_WRITE_reservd5	(0x0C)
#define NS8390_PAGE2_WRITE_reservd6	(0x0D)
#define NS8390_PAGE2_WRITE_reservd7	(0x0E)
#define NS8390_PAGE2_WRITE_reservd8	(0x0F)


/*
 * Bit pattern definitions for the chipset.
 */

/* Command Register, CR (rd wr) */
#define NS8390_CR_STP	(0x01)
#define NS8390_CR_STA	(0x02)
#define NS8390_CR_TXP	(0x04)
#define NS8390_CR_RD_RD	(0x08)
#define NS8390_CR_RD_WR	(0x10)
#define NS8390_CR_RD_SP	(0x18)
#define NS8390_CR_RD_AB	(0x20)
#define NS8390_CR_PAGE0	(0x00)
#define NS8390_CR_PAGE1	(0x40)
#define NS8390_CR_PAGE2	(0x80)
#define NS8390_CR_PAGE3	(0xC0)

/* Transmit Control Register, TCR (wr) */
#define NS8390_TCR_CRC		(0x01)
#define NS8390_TCR_LOOP_NO	(0x00)
#define NS8390_TCR_LOOP_I0	(0x02)
#define NS8390_TCR_LOOP_E1	(0x04)
#define NS8390_TCR_LOOP_E0	(0x06)
#define NS8390_TCR_ATD		(0x08)
#define NS8390_TCR_OFST		(0x10)

/* Data Configuration Register, DCR (wr) */
#define NS8390_DCR_WTS		(0x01)
#define NS8390_DCR_BOS		(0x02)
#define NS8390_DCR_LAS		(0x04)
#define NS8390_DCR_LS		(0x08)
#define NS8390_DCR_AR		(0x10)
#define NS8390_DCR_FT_2B	(0x00)
#define NS8390_DCR_FT_4B	(0x20)
#define NS8390_DCR_FT_8B	(0x40)
#define NS8390_DCR_FT_12B	(0x60)
#define NS8390_DCR_FT_1W	(0x00)
#define NS8390_DCR_FT_2W	(0x20)
#define NS8390_DCR_FT_4W	(0x40)
#define NS8390_DCR_FT_6W	(0x60)

/* Receive Configuration Register, RCR (wr) */
#define NS8390_RCR_SEP		(0x01)
#define NS8390_RCR_AR		(0x02)
#define NS8390_RCR_AB		(0x04)
#define NS8390_RCR_AM		(0x08)
#define NS8390_RCR_PRO		(0x10)
#define NS8390_RCR_MON		(0x20)

/* Interrupt Status Register, ISR (rd wr) */
#define NS8390_ISR_PRX		(0x01)
#define NS8390_ISR_PTX		(0x02)
#define NS8390_ISR_RXE		(0x04)
#define NS8390_ISR_TXE		(0x08)
#define NS8390_ISR_OVW		(0x10)
#define NS8390_ISR_CNT		(0x20)
#define NS8390_ISR_RDC		(0x40)
#define NS8390_ISR_RST		(0x80)

/* Interrupt Mask Register, IMR (wr) */
#define NS8390_IMR_PRXE		(0x01)
#define NS8390_IMR_PTXE		(0x02)
#define NS8390_IMR_RXEE		(0x04)
#define NS8390_IMR_TXEE		(0x08)
#define NS8390_IMR_OVWE		(0x10)
#define NS8390_IMR_CNTE		(0x20)
#define NS8390_IMR_RDCE		(0x40)

/* Received Status Register, RSR, and Packet Status Byte */
#define NS8390_RSR_PRX		(0x01)
#define NS8390_RSR_CRC		(0x02)
#define NS8390_RSR_FAE		(0x04)
#define NS8390_RSR_FO		(0x08)
#define NS8390_RSR_MPA		(0x10)
#define NS8390_RSR_PHY		(0x20)
#define NS8390_RSR_DIS		(0x40)
#define NS8390_RSR_DFR		(0x80)

/* Transmit Status Register, TSR (rd) */
#define NS8390_TSR_PTX		(0x01)
#define NS8390_TSR_COL		(0x04)
#define NS8390_TSR_ABT		(0x08)
#define NS8390_TSR_CRS		(0x10)
#define NS8390_TSR_FU		(0x20)
#define NS8390_TSR_CDH		(0x40)
#define NS8390_TSR_OWC		(0x80)

/*******************************************************************/
#endif /* _NS8390_H */

