/*
 * RTEMS/KA9Q driver for NE2000 network card in SBC5206 ISA slot
 *
 * Written by David Fiddes, D.J.Fiddes@hw.ac.uk
 * http://www.calm.hw.ac.uk/davidf/coldfire/
 *
 *  $Id:  $
 */
#include <bsp.h>
#include <rtems/error.h>
#include <ka9q/rtems_ka9q.h>
#include <ka9q/global.h>
#include <ka9q/enet.h>
#include <ka9q/iface.h>
#include <ka9q/netuser.h>
#include <ka9q/trace.h>
#include <ka9q/commands.h>
#include <ka9q/domain.h>
#include <ns8390.h>


/* 0 = no debug, 1 = standard traces, 2 = register accesses, 3 = all accesses inc data */
#define DEBUG_LEVEL		1

static const char *version = 
	"NE2000 RTEMS/KA9Q driver A0.01 02/08/98 David Fiddes (D.J@fiddes.surfaid.org)\n";


#define ISABASE (0x40000000)
#define WIERDOFFSET (0x10000)

typedef volatile struct
{
	unsigned16	_CR;
	union
	{
		struct
		{
			/* Even registers */
			unsigned16	CLDA1;		/* CLDA1 (rd) PSTOP (wr) */
			unsigned16	TSR;		/* TSR (rd) TPSR (wr) */
			unsigned16	FIFO;		/* FIFO (rd) TBCR1 (wr) */
			unsigned16	CRDA0;		/* CRDA0 (rd) RSAR0 (wr) */
			unsigned16	RBCR0;		/* Remote Byte Count 0 (wr) */
			unsigned16	RSR;		/* RSR (rd) RCR (wr) */
			unsigned16	CNTR1;		/* CNTR1 (rd) DCR (wr) */

			unsigned16	DATAPORT;

			unsigned16	reserved[(0x10000-0x0012)/2];

			/* Odd registers */
			unsigned16	CLDA0;		/* CLDA0 (rd) PSTART (wr) */
			unsigned16	BNRY;		/* Boundary pointer (rd wr) */
			unsigned16	NCR;		/* NCR (rd) TBCR0 (wr) */
			unsigned16	ISR;		/* Interrupt Status Register (rd wr) */
			unsigned16	CRDA1;		/* CRDA1 (rd) RSAR1 (wr) */
			unsigned16	RBCR1;		/* Remote Byte Count 1 (wr) */
			unsigned16	CNTR0;		/* CNTR0 (rd) TCR (wr) */
			unsigned16	CNTR2;		/* CNTR2 (rd) IMR (wr) */
		} page0;
		struct
		{
			/* Even registers */
			unsigned16	PAR1;		/* Physical Address Byte 1 */
			unsigned16	PAR3;		/* Physical Address Byte 3 */
			unsigned16	PAR5;		/* Physical Address Byte 5 */
			unsigned16	MAR0;		/* Multicast Address Byte 0 */
			unsigned16	MAR2;		/* Multicast Address Byte 2 */
			unsigned16	MAR4;		/* Multicast Address Byte 4 */
			unsigned16	MAR6;		/* Multicast Address Byte 6 */

			unsigned16	reserved[(0x10000-0x0010)/2];

			/* Odd registers */
			unsigned16	PAR0;		/* Physical Address Byte 0 */
			unsigned16	PAR2;		/* Physical Address Byte 2 */
			unsigned16	PAR4;		/* Physical Address Byte 4 */
			unsigned16	CURR;		/* Current Page Register (rd wr) */
			unsigned16	MAR1;		/* Multicast Address Byte 1 */
			unsigned16	MAR3;		/* Multicast Address Byte 3 */
			unsigned16	MAR5;		/* Multicast Address Byte 5 */
			unsigned16	MAR7;		/* Multicast Address Byte 7 */
		} page1;
		struct
		{
			/* Even registers */
			unsigned16	PSTOP;		/* PSTOP (rd) CLDA1 (wr) */
			unsigned16	TPSR;		/* Transmit Page Start Address (rd) */
			unsigned16	ACU;		/* Address Counter Upper */
			unsigned16	reserved0;
			unsigned16	reserved2;
			unsigned16	RCR;		/* Receive Configuration Register (rd) */
			unsigned16	DCR;		/* Data Configuration Register (rd) */

			unsigned16	reserved[(0x10000-0x0010)/2];

			/* Odd registers */
			unsigned16	PSTART;		/* PSTART (rd) CLDA0 (wr) */
			unsigned16	RNPP;		/* Remote Next Packet Pointer */
			unsigned16	LNPP;		/* Local Next Packet Pointer */
			unsigned16	ACL;		/* Address Counter Lower */
			unsigned16	reserved1;
			unsigned16	reserved3;
			unsigned16	TCR;		/* Transmit Configuration Register (rd) */
			unsigned16	IMR;		/* Interrupt Mask Register (rd) */
		} page2;
	} regs;
} NS8390;

#define NS8390_DATAPORT(PTR)	(PTR->regs.page0.DATAPORT)

	
	
/*
 * Module Prototypes
 */


/*
 * I/O function implementation
 */

static unsigned8
ne2000_getRegister( unsigned32 port, unsigned8 reg )
{
	volatile unsigned16 *pRegister;
	int					odd_addr;

	odd_addr = reg & 1;

	if ( odd_addr )
	{
		pRegister = (unsigned16 *)( ISABASE + WIERDOFFSET + port + reg - 1 );
	}
	else
	{
		pRegister = (unsigned16 *)( ISABASE + port + reg );
	}

#if DEBUG_LEVEL > 1    
    printf("ne2000_getRegister( port = %4.4X, reg = %X )\n", port, reg );
	printf("  pRegsiter = %4.4X\n", (unsigned32)pRegister );
#endif

	return *pRegister;
}


static void
ne2000_setRegister(unsigned32 port, unsigned8 reg, unsigned8 value)
{
	volatile unsigned16 *pRegister;
	int					odd_addr;

	odd_addr = reg & 1;

	if ( odd_addr )
	{
		pRegister = (unsigned16 *)( ISABASE + WIERDOFFSET + port + reg - 1 );
	}
	else
	{
		pRegister = (unsigned16 *)( ISABASE + port + reg );
	}

#if DEBUG_LEVEL > 1    
    printf("ne2000_setRegister( port = %4.4X, reg = %X, value = %X )\n", port, reg, value );
	printf("  pRegister = %4.4X\n", (unsigned32)pRegister );
#endif

	*pRegister = value;
}

#if 0
static unsigned16
ne2000_getData( unsigned32 port )
{
	volatile unsigned16 *pRegister;
	unsigned16			data;

	pRegister = (unsigned16 *)( ISABASE + port + NS8390_PAGE0_DATAPORT );

#if DEBUG_LEVEL > 2
    printf("ne2000_getData( port = %4.4X )\n", port );
	printf("  pRegister = %4.4X\n", (unsigned32)pRegister );
#endif

	data = *pRegister;

	printf( "ne2000_getData() returns %4.4X - honest\n", data );


	return data;
}

#endif

static unsigned16
ne2000_getData( unsigned32 port )
{
	NS8390				*ns8390;
	unsigned16			data;


	ns8390 = (NS8390 *)0x40000300;
	

	
	
	data = NS8390_DATAPORT(ns8390);

	printf( "ne2000_getData() returns %4.4X - honest\n", data );


	return data;
}




static void
ne2000_setData(unsigned32 port, unsigned16 value)
{
	volatile unsigned16 *pRegister;

	pRegister = (unsigned16 *)( ISABASE + port + NS8390_PAGE0_DATAPORT );

#if DEBUG_LEVEL > 2
    printf("ne2000_setData( port = %4.4X, value = %2.2X )\n", port, value );
	printf("  pRegister = %4.4X\n", (unsigned32)pRegister );
#endif

	*pRegister = value;
}


static void
ne2000_block_output( struct ns8390_device *dp, int *len, struct mbuf **bpp, int start_page )
{
	struct mbuf				*bp;
	uint8					*p;
	int						index;

	unsigned32              ne2000_port;

	setRegister_f           setReg;
	getRegister_f           getReg;
	setData_f				setData;

	printf( "ne2000_block_output() - start\n" );

	/* setup the IO functions */
	ne2000_port = dp->port;
	setReg		= dp->setRegister;
	getReg		= dp->getRegister;
	setData		= dp->setData;

	/* flip through the mbufs describing the packet to work out how big it is */
	bp = *bpp;
	*len = bp->cnt;
	while ( bp->next != NULL )
	{
		bp = bp->next;
		*len = *len + bp->cnt;
	}


	/* since we're a word output device upsize to an even number of bytes */
	if (*len & 0x01)
		*len = *len + 1;
	

	/* Make sure we're on PAGE0, abort any transfer and stay started */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_RD_AB | NS8390_CR_STA );


	/* set the size of the dma transfer */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RBCR0, *len & 0xFF );
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RBCR1, (*len >> 8) & 0xFF );

	/* set the source of the dma transfer */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RSAR0, 0x00 );
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RSAR0, start_page );

	/* start remote write */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_STA | NS8390_CR_RD_WR );

	printf( "About to spool the packet to the chip...\n" );

	/* spool each mbuf to the chip */
	bp = *bpp;
	do {
		/* output the actual data */
		p = bp->data;
		for (index = 0; index < bp->cnt; index = index + 2)
		{
			(*setData)( ne2000_port, *(uint16*)(p + index) );
		}
		
		bp = bp->next;
	} while ( bp != NULL );

	printf( "ne2000_block_output() - end\n" );

}

static void
ne2000_block_input( struct ns8390_device *dp, int len, struct mbuf *bp, int ring_offset )
{
	uint8					*p;
	int						index;

	unsigned32              ne2000_port;

	setRegister_f           setReg;
	getRegister_f           getReg;
	getData_f				getData;

	printf( "ne2000_block_input() - start\n" );

	/* setup the IO functions */
	ne2000_port = dp->port;
	setReg		= dp->setRegister;
	getReg		= dp->getRegister;
	getData		= dp->getData;

	/* Make sure we're on PAGE0, abort any transfer and stay started */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_RD_AB | NS8390_CR_STA );

	/* set the size of the dma transfer */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RBCR0, len & 0xFF );
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RBCR1, (len >> 8) & 0xFF );

	/* set the source of the dma transfer */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RSAR0, ring_offset & 0xFF );
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RSAR0, (ring_offset >> 8) & 0xFF );

	/* start remote read */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_STA | NS8390_CR_RD_RD );

	bp->cnt = len;

	p = bp->data;
	for (index = 0; index < len; index = index + 2 )
	{
		*(uint16*)(p + index) = (*getData)( ne2000_port );
	}

	printf( "ne2000_block_input - end\n" );
}


static void
ne2000_get_pkt_hdr( struct ns8390_device *dp, struct ns8390_pkt_hdr *header, int ring_page )
{
	unsigned32              ne2000_port;

	setRegister_f           setReg;
	getRegister_f           getReg;
	getData_f				getData;
	uint16					a;
	uint16					b;

	printf( "ne2000_get_pkt_hdr() - start\n" );

	/* setup the IO functions */
	ne2000_port = dp->port;
	setReg		= dp->setRegister;
	getReg		= dp->getRegister;
	getData		= dp->getData;

	/* Make sure we're on PAGE0, abort any transfer and stay started */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_RD_AB | NS8390_CR_STA );

	/* set the size of the dma transfer */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RBCR0, 4 );
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RBCR1, 0 );

	/* set the source of the dma transfer */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RSAR0, 0 );
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RSAR0, ring_page );

	/* start remote read */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_STA | NS8390_CR_RD_RD );


	a = (*getData)( ne2000_port );
	b = (*getData)( ne2000_port );


	printf( "Hmm  a = %.4X, b = %.4X\n", a, b );



/*	header->status = a & 0xFF;
	header->next = (a >> 8) & 0xFF;

	header->count = ((b & 0xFF) << 8) | ((b >> 8) & 0xFF);  */

	header->status = (a >> 8) & 0xFF;
	header->next   = a & 0xFF;

	header->count = b;

	printf( "ne2000_get_pkt_hdr() - end\n" );
}


static void
ne2000_get_eth_addr( struct ns8390_device *dp, uint8 *hwa )
{
	int						index;
	unsigned8				eeprom[32];
	unsigned32              ne2000_port;

	setRegister_f           setReg;
	getRegister_f           getReg;
	getData_f				getData;


	printf( "ne2000_get_eth_addr() - start\n" );

	/* setup the IO functions */
	ne2000_port = dp->port;
	setReg		= dp->setRegister;
	getReg		= dp->getRegister;
	getData		= dp->getData;


	/* read the ethernet address from eprom ? */

	/* Make sure we are on PAGE0, abort any current DMA transfers and stop the chip */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_RD_AB | NS8390_CR_STP );


	/* set FIFO Threshold to 8bytes */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_DCR, NS8390_DCR_FT_8B );

	/* Reset the dma transfer size */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RBCR0, 0 );
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RBCR1, 0 );

	/* mask and clear all the chip's interrupts */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_IMR, 0 );
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_ISR, 0xFF );

	/* Set to do an internal loopback *//* MON seems undocumented */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RCR, NS8390_RCR_MON );
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_TCR, NS8390_TCR_LOOP_I0 );

	/* set transfer size to 32 bytes(sizeof EEPROM) */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RBCR0, 32 );
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RBCR1, 0 );
	
	/* make transfer go to 0 */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RSAR0, 0 );
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RSAR1, 0 );

	/* start the transfer */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_STA | NS8390_CR_RD_RD );

	/* load in the EEPROM data */
	/* NE2000s double up the PROM data which is why this works */
	for (index = 0; index < 16; ++index)
		eeprom[index] = (*getData)( ne2000_port );

	/* suck out the Ethernet address that we need */
	for (index = 0; index < 6; index++)
		hwa[index] = eeprom[index*2]; 

	/* Abort any transfer and stop the chip */
	(*setReg)( ne2000_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_STP | NS8390_CR_RD_AB );

	printf( "ne2000_get_eth_addr() - end\n" );

}



/*
 * Attach an NE2000 driver to the system
 * This is the only `extern' function in the driver.
 *
 * argv[0]: interface label, e.g., "rtems"
 * The remainder of the arguemnts are key/value pairs:
 * mtu ##                  --  maximum transmission unit, default 1500
 * ip ###.###.###.###      -- IP address
 * ether ##:##:##:##:##:## -- Ethernet address
 * ether eeprom            -- Get Ethernet address from the ISA card's internal EEPROM
 */
int
rtems_ka9q_driver_attach (int argc, char *argv[], void *p)
{
	int						argIndex;
 	struct ns8390_device	*dp;
	struct iface			*iface;
	int						status;
	char					cbuf[30];
	rtems_status_code		sc;
	rtems_isr_entry			old_handler;

	/* Print module version information */
	printf( version );

	/* dump the parameters */
	printf("rtems_ka9q_driver_attach() called with the following parameters\n");
	for (argIndex = 0 ; argIndex < argc ; argIndex++)
	{
		printf( "argv[%d] = %s\n", argIndex, argv[argIndex] );
	}


	printf("\n");
	
	/* Does an interface of this name already exist? */
	if (if_lookup (argv[0]) != NULL)
	{
		printf ("Interface %s already exists\n", argv[0]);
		return -1;
	}
 

	/* Carry out generic ns8390 device initialisation */
	status = ns8390_driver_attach( &dp );

	if ( status < 0 )
	{
		return status;
	}

	/* Configure the NE2000 ns8390 configuration */

	/* Setup the device specific I/O functions */
	dp->getRegister = &ne2000_getRegister;
	dp->setRegister = &ne2000_setRegister;
	dp->getData	    = &ne2000_getData;
	dp->setData     = &ne2000_setData;

	dp->block_output   = &ne2000_block_output;
	dp->block_input    = &ne2000_block_input;
	dp->get_ns8390_hdr = &ne2000_get_pkt_hdr;

	/* manually set the port for now */
	dp->port = 0x300;

	/* Setup the default internal memory configuration for the NS8390 */
	/* TODO: This code assumes that we only talk to a 16bit card */
	dp->tx_start_page = 0x40;
	dp->stop_page     = 0x80;
	dp->rx_start_page = dp->tx_start_page + NUM_TX_PAGES;


	/* grab a pointer to the interface */
	iface = dp->iface;

	/* give it our name */
	iface->name = strdup (argv[0]);

	/*
	 * Parse arguments
	 */
	for (argIndex = 1 ; argIndex < (argc - 1) ; argIndex++) {
		if (strcmp ("mtu", argv[argIndex]) == 0)
		{
			iface->mtu = atoi (argv[++argIndex]);
		}
		else if (strcmp ("ip", argv[argIndex]) == 0) 
		{
			iface->addr = resolve (argv[++argIndex]);
		}
		else if (strcmp ("ether", argv[argIndex]) == 0) 
		{
			argIndex++;
			if (strcmp (argv[argIndex], "eeprom") == 0) 
			{
					ne2000_get_eth_addr( dp, iface->hwaddr );
			}
			else 
			{
				gether (iface->hwaddr, argv[argIndex]);
			}
		}
		else
		{
			printf ("Argument %d (%s) is invalid.\n", argIndex, argv[argIndex]);
			return -1;
		}
	}

	printf ("Ethernet address: %s\n", pether (cbuf, iface->hwaddr));
	printf ("Internet address: %s\n", inet_ntoa(iface->addr));
	printf ("Irqno: %X, port: %X\n",dp->irno, dp->port);
	fflush(stdout);


	/*
	 * Set up NS8390 hardware
	 */

	ns8390_init( dp );

	/* Setup ColdFire interrupts */
	sc = rtems_interrupt_catch(	ns8390_interrupt_handler,
								NIC_IRQ_VECTOR,
								&old_handler );
	if (sc != RTEMS_SUCCESSFUL)
		rtems_panic ("Can't attach 68HC901(NE2000) interrupt handler: %s\n",
  						rtems_status_text (sc));
	sim->icr4 =	MCF5206_SIM_ICR_IL(NIC_IRQ_LEVEL) |
				MCF5206_SIM_ICR_IP(NIC_IRQ_PRIORITY);
	sim->imr = sim->imr & (~MCF5206_SIM_IMR_EINT4);	/* Enable 68HC901 interrupts */


	/* Add the network interface to the system and start the chip */
	ns8390_start( dp );


	printf( "ne2000_driver_attach() - end\n");

	return 0;
}
